#include "globals.h"

#include <fcntl.h>

#ifdef USE_PTY
#include <grp.h>
#include <termios.h>

#ifndef	TIOCGWINSZ
#include <sys/ioctl.h>	/* 44BSD requires this */
#endif

#ifdef SYSV
#include <sys/wait.h>
#endif
#endif

static struct Children *children = NULL;
static Widget dialog = NULL, cmdTF, runInXtermToggle;

void getChildMessages( 
#if NeedFunctionPrototypes
	XtPointer, int*, XtInputId* 
#endif
);
void allocateChild( 
#if NeedFunctionPrototypes
	XtInputId, char*, pid_t 
#endif
);
#ifdef USE_PTY
int ptym_open( 
#if NeedFunctionPrototypes
	char* 
#endif
);
int ptys_open( 
#if NeedFunctionPrototypes
	int, char* 
#endif
);
#endif
void handleSignal( 
#if NeedFunctionPrototypes
	XtInputId 
#endif
);
void cmdOKCB( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);
void cmdCancelCB( 
#if NeedFunctionPrototypes
	Widget, XtPointer, XtPointer 
#endif
);


#ifdef USE_PTY
int ptym_open( pts_name )
  char *pts_name;
{
  int fdm;
  char *ptr1, *ptr2;

  strcpy(pts_name, "/dev/ptyXY");
  /* array index: 0123456789 (for references in following code) */
  for( ptr1 = "pqrstuvwxyzPQRST"; *ptr1 != 0; ptr1++ ) 
  {
    pts_name[8] = *ptr1;
    for( ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++ ) 
    {
      pts_name[9] = *ptr2;

      /* try to open master */
      if( (fdm = open(pts_name, O_RDWR)) < 0 ) 
      {
	if( errno == ENOENT )	/* different from EIO */
	    return( -1 );	/* out of pty devices */
	else
	    continue;		/* try next pty device */
      }

      pts_name[5] = 't';	/* change "pty" to "tty" */
      return( fdm );		/* got it, return fd of master */
    }
  }

  return( -1 );			/* out of pty devices */
}

int ptys_open( fdm, pts_name )
  int fdm;
  char *pts_name;
{
  struct group *grptr;
  int gid, fds;

  if ( (grptr = getgrnam("tty")) != NULL )
    gid = grptr->gr_gid;
  else
    gid = -1;		/* group tty is not in the group file */

  /* following two functions don't work unless we're root */
  chown( pts_name, getuid(), gid );
  chmod( pts_name, S_IRUSR | S_IWUSR | S_IWGRP );

  if ( (fds = open(pts_name, O_RDWR)) < 0 ) 
  {
    close( fdm );
    return( -1 );
  }

  return( fds );
}
#endif

void launchCmd( w, cmd )
  Widget w;
  char *cmd;
{
  int  fdm, fifo[2];
  pid_t pid;
  char  pts_name[20];
  XtInputId id;

#ifdef DEBUG
  printf( "Attempting to launch command = %s\n", cmd );
#endif

#ifdef USE_PTY
  if( (fdm = ptym_open(pts_name)) < 0 )
  {
    popupErrorDialog( "Can't open master pseudo terminal..." );
    return;
  }
  id = XtAppAddInput( XtWidgetToApplicationContext(w), fdm, 
	(XtPointer)XtInputReadMask, getChildMessages, (XtPointer)NULL );
#else
  if( pipe( fifo ) == -1 )
  {
    popupErrorDialog( "pipe() failed..." );
    return;
  }
  id = XtAppAddInput( XtWidgetToApplicationContext(w), fifo[0],
        (XtPointer)XtInputReadMask, getChildMessages, (XtPointer)NULL );
#endif

  fcntl( ConnectionNumber(XtDisplay(w)), F_SETFD, 1 );
  
  if( (pid = fork()) < 0 )
  {
    popupErrorDialog( "fork() failed..." );
    return;
  }
  else if( pid == 0 ) 
  {
    char shell[256], *shellBase;
    int fds;

    strcpy( shell, "" );
    strcpy( shell, getenv("SHELL") );
    if( !strcmp( shell, "" ) )  strcpy( shell, "/bin/sh" );
    shellBase = strrchr( shell, '/' ) + 1;

#ifdef USE_PTY
    if( setsid() < 0 )
    {
      fprintf( stderr, "***MFM ERROR: setsid() failed...\n" );
      return;
    }

    /* SVR4 acquires controlling terminal on open() */
    if( (fds = ptys_open(fdm, pts_name)) < 0 )
    {
      fprintf( stderr, "***MFM ERROR: Can't open slave pseudo terminal...\n" );
      return;
    }

    /* all done with master in child */
    close( fdm );

#if defined(TIOCSCTTY) && !defined(CIBAUD)
    /* 44BSD way to acquire controlling terminal */
    /* !CIBAUD to avoid doing this under SunOS   */
    if( ioctl(fds, TIOCSCTTY, (char *)0) < 0 )
    {
      fprintf( stderr, "***MFM ERROR: ioctl() TIOCSCTTY error...\n" );
      return;
    }
#endif

    if( dup2(fds, fileno(stdin)) != fileno(stdin) )
    {
      fprintf( stderr, "***MFM ERROR: dup2 error to stdin\n" );
      return;
    }
    if( dup2(fds, fileno(stdout)) != fileno(stdout) )
    {
      fprintf( stderr, "***MFM ERROR: dup2 error to stdout\n" );
      return;
    }
    if( dup2(fds, fileno(stderr)) != fileno(stderr) )
    {
      fprintf( stderr, "***MFM ERROR: dup2 error to stderr\n" );
      return;
    }
    if( fds > fileno(stderr) ) close( fds );
#else
    freopen( "/dev/console", "a", stdout );
    freopen( "/dev/console", "a", stderr );
    freopen( "/dev/console", "a", stdin );
#endif

#ifdef DEBUG
    printf( "Executing: %s -c '%s'\n", shell, cmd );
#endif

    execl( shell, shellBase, "-c", cmd, NULL );
    fprintf( stderr, "***MFM ERROR: execl() failed for: %s -c '%s'\n", shell, cmd );
    exit( -1 );
  } 
  else
  {
#ifndef USE_PTY
    close( fifo[1] );
#endif
    allocateChild( id, cmd, pid );
  }
}

void allocateChild( id, cmd, pid )
  XtInputId id;
  char *cmd;
  pid_t pid;
{
  struct Children *newentry = appendChild( &children );
  newentry->id = id;
  newentry->pid = pid;
  newentry->cmd = (char *)XtMalloc( strlen(cmd)+1 );
  strcpy( newentry->cmd, cmd );
}

void getChildMessages( w, source, id )
  XtPointer w;
  int* source;
  XtInputId* id;
{
  char buffer[1056];
  int n;

  memset( (void *)buffer, '\0', sizeof(buffer) );

  if( (n = read( *source, buffer, sizeof(buffer)-1 )) > 0 ) 
  {
    buffer[n] = '\0';
    popupErrorDialog( buffer );
  }
  else if( n < 0 )
  {
    XtRemoveInput( *id );
    handleSignal( *id );
  }
}

void handleSignal( id )
  XtInputId id;
{
  int i, l = numChildren( children );
  for( i = 0; i < l; i++ )
  {
    struct Children *child = getIthChild( children, i );
    if( id == child->id )
    {
      removeIthChild( &children, i );
      break;
    }
  }
}

void popupCmdDialog( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  if( !dialog )
  {
    Widget topForm, separator, okButton, cancelButton, cmdForm, cmdLabel;
    Atom wmDel;

    dialog = XtVaCreatePopupShell( "Execute Command",
	transientShellWidgetClass, w,
	XmNwidth, 450,
        XmNheight, 150,
        NULL );

    topForm = XtVaCreateManagedWidget( "topForm",
	xmFormWidgetClass, dialog, NULL );

    XtVaSetValues( dialog, XmNdeleteResponse, XmDO_NOTHING, NULL );
    wmDel = XmInternAtom( XtDisplay(topForm), "WM_DELETE_WINDOW", False );
    XmAddWMProtocolCallback( dialog, wmDel, cmdCancelCB, topForm );

    separator = XtVaCreateManagedWidget( "separator",
                xmSeparatorWidgetClass, topForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 60,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                NULL );

    okButton = XtVaCreateManagedWidget( "Execute",
                xmPushButtonWidgetClass, topForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 10,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 15,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 35,
                XmNheight, 30,
                XmNshowAsDefault, True,
                NULL );
    XtAddCallback( okButton, XmNactivateCallback, cmdOKCB, NULL );

    cancelButton = XtVaCreateManagedWidget( "Cancel",
                xmPushButtonWidgetClass, topForm,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 15,
                XmNleftAttachment, XmATTACH_POSITION,
                XmNleftPosition, 65,
                XmNrightAttachment, XmATTACH_POSITION,
                XmNrightPosition, 85,
                XmNheight, 30,
                NULL );
    XtAddCallback( cancelButton, XmNactivateCallback, cmdCancelCB, NULL );

    cmdForm = XtVaCreateManagedWidget( "cmdForm",
                xmFormWidgetClass, topForm,
                XmNtopAttachment, XmATTACH_FORM,
                XmNtopOffset, 20,
                XmNleftAttachment, XmATTACH_FORM,
                XmNleftOffset, 10,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 10,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 50,
                NULL );

    runInXtermToggle = XtVaCreateManagedWidget( "Run In XTerm",
                xmToggleButtonWidgetClass, cmdForm,
		XmNtopAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                NULL );

    cmdTF = XtVaCreateManagedWidget( "cmdTF",
		xmTextFieldWidgetClass, cmdForm,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, runInXtermToggle,
		XmNleftAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		XmNmarginHeight, 2,
		XmNheight, 30,
                NULL );

    cmdLabel = XtVaCreateManagedWidget( "Command:",
                xmLabelWidgetClass, cmdForm,
		XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, cmdTF,
                XmNleftAttachment, XmATTACH_FORM,
                NULL );
  }

  XtPopup( dialog, XtGrabNone );
  XRaiseWindow( XtDisplay(dialog), XtWindow(dialog) );
}

void cmdOKCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  char *cmd, *txt = XmTextFieldGetString( cmdTF );

  if( XmToggleButtonGetState(runInXtermToggle) )
  {
    cmd = (char *)XtMalloc( 2*strlen(txt) + 50 );
    sprintf( cmd, "xterm -title '%s' -e %s '%s'", txt, 
	createPausemeFile(), txt );
  }
  else
  {
    cmd = (char *)XtMalloc( strlen(txt)+5 );
    strcpy( cmd, txt );
  }

  launchCmd( w, cmd );
  XtFree( (char *)txt );
  XtFree( (char *)cmd );
  XmTextFieldSetString( cmdTF, "" );
  XtPopdown( dialog );
}

void cmdCancelCB( w, x1, x2 )
  Widget w;
  XtPointer x1, x2;
{
  XmTextFieldSetString( cmdTF, "" );
  XtPopdown( dialog );
}

void popupCmdWithCmd( w, s )
  Widget w;
  char *s;
{
  popupCmdDialog( w, (XtPointer)NULL, (XtPointer)NULL );
  XmTextFieldSetString( cmdTF, s );
}
